@@ -1,5 +1,8 @@
 import json
+from collections import namedtuple
 from snappi_ixnetwork.configurebgp import ConfigureBgp
+from snappi_ixnetwork.deviceCompactor import DeviceCompactor
+from snappi_ixnetwork.timer import Timer
 
 
 class Ngpf(object):
@@ -11,11 +14,11 @@ class Ngpf(object):
     """
 
     _TPID_MAP = {
-        "8100": "ethertype8100",
-        "88a8": "ethertype88a8",
-        "9100": "ethertype9100",
-        "9200": "ethertype9200",
-        "9300": "ethertype9300",
+        "x8100": "ethertype8100",
+        "x88a8": "ethertype88a8",
+        "x9100": "ethertype9100",
+        "x9200": "ethertype9200",
+        "x9300": "ethertype9300",
     }
 
     # Select type of Traffic
@@ -60,6 +63,23 @@ class Ngpf(object):
             return len(errata) == 0
         return True
 
+    def _get_devices_info(self, devices_in_topo):
+        DeviceInfo = namedtuple("DeviceInfo", ["device", "multiplier"])
+        dev_info_list = []
+        if self._api.do_compact is True:
+            with Timer(self._api, "Compacting snappi objects :"):
+                sim_dev_list = DeviceCompactor(devices_in_topo).compact()
+                for sim_div in sim_dev_list:
+                    if sim_div.len > 1:
+                        self._api.set_dev_compacted(sim_div.compact_dev)
+                    dev_info_list.append(
+                        DeviceInfo(sim_div.compact_dev, sim_div.len)
+                    )
+        else:
+            for dev in devices_in_topo:
+                dev_info_list.append(DeviceInfo(dev, 1))
+        return dev_info_list
+
     def _configure_topology(self, ixn_topology, devices):
         """One /topology for every unique device.container_name
         Topology name is device.container_name
@@ -80,48 +100,56 @@ class Ngpf(object):
         self._api._remove(ixn_topology, topologies.values())
         for topo_name, devices_in_topo in devices_in_topos.items():
             ixn_topology.find(Name="^%s$" % self._api.special_char(topo_name))
+            dev_info_list = self._get_devices_info(devices_in_topo)
+            cmt_devices = [dev_inf.device for dev_inf in dev_info_list]
             if len(ixn_topology) > 0:
-                self._api._remove(ixn_topology.DeviceGroup, devices_in_topo)
-        for device in devices:
-            args = {
-                "Name": self._api._get_topology_name(device.container_name),
-                "Ports": [self._api.ixn_objects[device.container_name]],
-            }
-            ixn_topology.find(
-                Name="^%s$" % self._api.special_char(args["Name"])
-            )
-            if len(ixn_topology) == 0:
-                ixn_topology.add(**args)
-            else:
-                self.update(ixn_topology, **args)
-            self._api.ixn_objects[ixn_topology.Name] = ixn_topology.href
-            self._configure_device_group(ixn_topology.DeviceGroup, device)
+                self._api._remove(ixn_topology.DeviceGroup, cmt_devices)
+            for device_info in dev_info_list:
+                device = device_info.device
+                multiplier = device_info.multiplier
+                container_name = device.get("container_name")
+                args = {
+                    "Name": self._api._get_topology_name(container_name),
+                    "Ports": [self._api.get_ixn_href(container_name)],
+                }
+                ixn_topology.find(
+                    Name="^%s$" % self._api.special_char(args["Name"])
+                )
+                if len(ixn_topology) == 0:
+                    ixn_topology.add(**args)
+                else:
+                    self.update(ixn_topology, **args)
+                self._api.set_ixn_object(ixn_topology.Name, ixn_topology.href)
+                self._configure_device_group(
+                    ixn_topology.DeviceGroup, device, multiplier
+                )
 
-    def _configure_device_group(self, ixn_device_group, device):
+    def _configure_device_group(self, ixn_device_group, device, multiplier):
         """Transform /components/schemas/Device into /topology/deviceGroup
         One /topology/deviceGroup for every device in port.devices
         """
-        args = {"Name": device.name, "Multiplier": 1}
-        ixn_device_group.find(
-            Name="^%s$" % self._api.special_char(device.name)
-        )
+        name = device.get("name")
+        args = {"Name": name, "Multiplier": multiplier}
+        ixn_device_group.find(Name="^%s$" % self._api.special_char(name))
         if len(ixn_device_group) == 0:
             ixn_device_group.add(**args)[-1]
         else:
             ixn_ng = ixn_device_group.NetworkGroup
             self._api._remove(ixn_ng, [])
             self.update(ixn_device_group, **args)
+        dg_href = ixn_device_group.href
+        self._api.set_ixn_cmp_object(device, dg_href, self.get_xpath(dg_href))
         self._config_proto_stack(ixn_device_group, device, ixn_device_group)
 
     def _config_proto_stack(self, ixn_obj, snappi_obj, ixn_dg):
-        self._api.ixn_objects[snappi_obj.name] = ixn_obj.href
-        properties = snappi_obj._properties
-        for prop_name in properties:
+        if not isinstance(snappi_obj, dict):
+            snappi_obj = snappi_obj._properties
+        for prop_name in snappi_obj.keys():
             stack_class = getattr(
                 self, "_configure_{0}".format(prop_name), None
             )
             if stack_class is not None:
-                child = properties[prop_name]
+                child = snappi_obj[prop_name]
                 if prop_name not in Ngpf._DEVICE_ENCAP_MAP:
                     raise Exception(
                         "Mapping is missing for {0}".format(prop_name)
@@ -129,10 +157,9 @@ class Ngpf(object):
                 self._api._device_encap[ixn_dg.Name] = Ngpf._DEVICE_ENCAP_MAP[
                     prop_name
                 ]
-                if child.name is not None:
-                    self._api._device_encap[
-                        child.name
-                    ] = Ngpf._DEVICE_ENCAP_MAP[prop_name]
+                child_name = child.get("name")
+                if child_name is not None:
+                    self._api.set_device_encap(child, Ngpf._DEVICE_ENCAP_MAP[prop_name])
                 new_ixn_obj = stack_class(ixn_obj, child, ixn_dg)
                 self._config_proto_stack(new_ixn_obj, child, ixn_dg)
 
@@ -168,23 +195,29 @@ class Ngpf(object):
             elif pattern.get("choice") == "random":
                 pass
 
-    def configure_value(self, source, attribute, value, enum_map=None):
+    def configure_value(
+        self, source, attribute, value, enum_map=None, multiplier=1
+    ):
         if value is None:
             return
         xpath = "/multivalue[@source = '{0} {1}']".format(source, attribute)
+        if multiplier > 1 and isinstance(value, list):
+            val_list = []
+            for val in value:
+                val_list.extend([val] * multiplier)
+            value = val_list
+        if isinstance(value, list) and len(set(value)) == 1:
+            value = value[0]
         if enum_map is not None:
-            value = enum_map[value]
-        if isinstance(value, list):
-            if len(set(value)) == 1:
-                ixn_value = {
-                    "xpath": "{0}/singleValue".format(xpath),
-                    "value": value[0],
-                }
+            if isinstance(value, list):
+                value = [enum_map[val] for val in value]
             else:
-                ixn_value = {
-                    "xpath": "{0}/valueList".format(xpath),
-                    "values": value,
-                }
+                value = enum_map[value]
+        if isinstance(value, list):
+            ixn_value = {
+                "xpath": "{0}/valueList".format(xpath),
+                "values": value,
+            }
         else:
             ixn_value = {
                 "xpath": "{0}/singleValue".format(xpath),
@@ -254,24 +287,27 @@ class Ngpf(object):
         ixn_ethernet = ixn_parent.Ethernet
         self._api._remove(ixn_ethernet, [ethernet])
         args = {}
-        ixn_ethernet.find(Name="^%s$" % self._api.special_char(ethernet.name))
+        eth_name = ethernet.get("name")
+        ixn_ethernet.find(Name="^%s$" % self._api.special_char(eth_name))
         if len(ixn_ethernet) == 0:
             ixn_ethernet.add(**args)
         else:
             self.update(ixn_ethernet, **args)
-        if ethernet.name is not None:
-            self._api.ixn_objects[ethernet.name] = ixn_ethernet.href
-            ixn_ethernet.Name = ethernet.name
+        if eth_name is not None:
+            ixn_ethernet.Name = eth_name
         eth_info = self.select_node(
             ixn_ethernet.href, children=["ipv4", "ipv6"]
         )
         eth_xpath = eth_info["xpath"]
-        self.configure_value(eth_xpath, "mac", ethernet.mac)
-        self.configure_value(eth_xpath, "mtu", ethernet.mtu)
-        if len(ethernet.vlans) > 0:
-            ixn_ethernet.VlanCount = len(ethernet.vlans)
+        self._api.set_ixn_cmp_object(ethernet, ixn_ethernet.href, eth_xpath)
+        
+        self.configure_value(eth_xpath, "mac", ethernet.get("mac"))
+        self.configure_value(eth_xpath, "mtu", ethernet.get("mtu"))
+        vlans = ethernet.get("vlans")
+        if vlans is not None and len(vlans) > 0:
+            ixn_ethernet.VlanCount = len(vlans)
             ixn_ethernet.EnableVlans.Single(ixn_ethernet.VlanCount > 0)
-            self._configure_vlan(ixn_ethernet.Vlan, ethernet.vlans)
+            self._configure_vlan(ixn_ethernet.Vlan, vlans)
         if (
             ethernet.get("ipv4") is not None
             and ethernet.get("ipv6") is not None
@@ -293,15 +329,15 @@ class Ngpf(object):
         """Transform Device.Vlan to /topology/.../vlan"""
         for i in range(0, len(ixn_vlans.find())):
             ixn_vlan = ixn_vlans[i]
-            if vlans[i].name is not None:
-                args = {"Name": vlans[i].name}
+            name = vlans[i].get("name")
+            if name is not None:
+                args = {"Name": name}
                 self.update(ixn_vlan, **args)
-                self._api.ixn_objects[vlans[i].name] = ixn_vlan.href
             vlan_xpath = self.get_xpath(ixn_vlan.href)
-            self.configure_value(vlan_xpath, "vlanId", vlans[i].id)
-            self.configure_value(vlan_xpath, "priority", vlans[i].priority)
+            self.configure_value(vlan_xpath, "vlanId", vlans[i].get("id"))
+            self.configure_value(vlan_xpath, "priority", vlans[i].get("priority"))
             self.configure_value(
-                vlan_xpath, "tpid", vlans[i].tpid, enum_map=Ngpf._TPID_MAP
+                vlan_xpath, "tpid", vlans[i].get("tpid"), enum_map=Ngpf._TPID_MAP
             )
 
     def _configure_ipv4(self, ixn_parent, ipv4, ixn_dg):
@@ -309,18 +345,19 @@ class Ngpf(object):
         ixn_ipv4 = ixn_parent.Ipv4
         self._api._remove(ixn_ipv4, [ipv4])
         args = {}
-        ixn_ipv4.find(Name="^%s$" % self._api.special_char(ipv4.name))
+        name = ipv4.get("name")
+        ixn_ipv4.find(Name="^%s$" % self._api.special_char(name))
         if len(ixn_ipv4) == 0:
             ixn_ipv4.add(**args)[-1]
         else:
             self.update(ixn_ipv4, **args)
-        if ipv4.name is not None:
-            ixn_ipv4.Name = ipv4.name
-            self._api.ixn_objects[ipv4.name] = ixn_ipv4.href
+        if name is not None:
+            ixn_ipv4.Name = name
         ip_xpath = self.get_xpath(ixn_ipv4.href)
-        self.configure_value(ip_xpath, "address", ipv4.address)
-        self.configure_value(ip_xpath, "gatewayIp", ipv4.gateway)
-        self.configure_value(ip_xpath, "prefix", ipv4.prefix)
+        self._api.set_ixn_cmp_object(ipv4, ixn_ipv4.href, ip_xpath)
+        self.configure_value(ip_xpath, "address", ipv4.get("address"))
+        self.configure_value(ip_xpath, "gatewayIp", ipv4.get("gateway"))
+        self.configure_value(ip_xpath, "prefix", ipv4.get("prefix"))
         return ixn_ipv4
 
     def _configure_bgpv4(self, ixn_parent, bgpv4, ixn_dg):
@@ -337,8 +374,8 @@ class Ngpf(object):
             self.update(ixn_ipv6, **args)
         if ipv6.name is not None:
             ixn_ipv6.Name = ipv6.name
-            self._api.ixn_objects[ipv6.name] = ixn_ipv6.href
         ip_xpath = self.get_xpath(ixn_ipv6.href)
+        self._api.set_ixn_cmp_object(ipv6, ixn_ipv6.href, ip_xpath)
         self.configure_value(ip_xpath, "address", ipv6.address)
         self.configure_value(ip_xpath, "gatewayIp", ipv6.gateway)
         self.configure_value(ip_xpath, "prefix", ipv6.prefix)
@@ -353,8 +390,31 @@ class Ngpf(object):
         names = payload.names
         if len(names) == 0:
             names = self._api.ixn_route_objects.keys()
+        ixn_obj_idx_list = {}
+        names = list(set(names))
         for name in names:
-            ixn_route = self._api.get_route_object(name)
-            ixn_route.Active.Single(Ngpf._ROUTE_STATE[payload.state])
+            route_info = self._api.get_route_object(name)
+            ixn_obj = None
+            for obj, index_list in ixn_obj_idx_list.items():
+                if obj.href == route_info.ixn_obj.href:
+                    ixn_obj = obj
+                    break
+            if ixn_obj is None:
+                ixn_obj_idx_list[route_info.ixn_obj] = list(range(
+                    route_info.index, route_info.index + route_info.multiplier
+                ))
+            else:
+                ixn_obj_idx_list[ixn_obj].extend(list(range(
+                    route_info.index, route_info.index + route_info.multiplier
+                )))
+        for obj, index_list in ixn_obj_idx_list.items():
+            index_list = list(set(index_list))
+            if len(index_list) == obj.Count:
+                obj.Active.Single(Ngpf._ROUTE_STATE[payload.state])
+            else:
+                values = obj.Active.Values
+                for idx in index_list:
+                    values[idx] = Ngpf._ROUTE_STATE[payload.state]
+                obj.Active.ValueList(values)
         self._api._ixnetwork.Globals.Topology.ApplyOnTheFly()
         return names
